#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import glob as globlib
import logging
import os
import sys
import time

import math
import numpy as np
import pandas as pd
import requests
from bravado.exception import BravadoConnectionError, BravadoTimeoutError, HTTPForbidden, \
    HTTPInternalServerError, HTTPServerError, HTTPUnauthorized, HTTPServiceUnavailable, HTTPRequestTimeout, \
    HTTPGatewayTimeout

from neptune.api_exceptions import ConnectionLost, Forbidden, ServerError, \
    Unauthorized, SSLError
from neptune.exceptions import InvalidNotebookPath, FileNotFound, NotAFile
from neptune.git_info import GitInfo

_logger = logging.getLogger(__name__)

IS_WINDOWS = hasattr(sys, 'getwindowsversion')


def map_values(f_value, dictionary):
    return dict(
        (k, f_value(v)) for k, v in dictionary.items()
    )


def map_keys(f_key, dictionary):
    return dict(
        (f_key(k), v) for k, v in dictionary.items()
    )


def as_list(value):
    if value is None or isinstance(value, list):
        return value
    else:
        return [value]


def validate_notebook_path(path):
    if not path.endswith(".ipynb"):
        raise InvalidNotebookPath(path)

    if not os.path.exists(path):
        raise FileNotFound(path)

    if not os.path.isfile(path):
        raise NotAFile(path)


def align_channels_on_x(dataframe):
    channel_dfs, common_x = _split_df_by_stems(dataframe)
    return merge_dataframes([common_x] + channel_dfs, on='x', how='outer')


def get_channel_name_stems(columns):
    return list(set([col[2:] for col in columns]))


def merge_dataframes(dataframes, on, how='outer'):
    merged_df = functools.reduce(lambda left, right: \
                                     pd.merge(left, right, on=on, how=how), dataframes)
    return merged_df


def is_float(value):
    try:
        _ = float(value)
    except ValueError:
        return False
    else:
        return True


def is_nan_or_inf(value):
    return math.isnan(value) or math.isinf(value)


def file_contains(filename, text):
    for line in open(filename):
        if text in line:
            return True
    return False


def in_docker():
    cgroup_file = '/proc/self/cgroup'
    return os.path.exists('./dockerenv') or (os.path.exists(cgroup_file) and file_contains(cgroup_file, text='docker'))


def is_notebook():
    try:
        # pylint: disable=pointless-statement,undefined-variable
        get_ipython
        return True
    except Exception:
        return False


def _split_df_by_stems(df):
    channel_dfs, x_vals = [], []
    for stem in get_channel_name_stems(df.columns):
        channel_df = df[['x_{}'.format(stem), 'y_{}'.format(stem)]]
        channel_df.columns = ['x', stem]
        channel_df = channel_df.dropna()
        channel_dfs.append(channel_df)
        x_vals.extend(channel_df['x'].tolist())
    common_x = pd.DataFrame({'x': np.unique(x_vals)}, dtype=float)
    return channel_dfs, common_x


def discover_git_repo_location():
    # pylint:disable=bad-option-value,import-outside-toplevel
    import __main__

    if hasattr(__main__, '__file__'):
        return os.path.dirname(os.path.abspath(__main__.__file__))
    return None


def update_session_proxies(session, proxies):
    if proxies is not None:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError("Wrong proxies format: {}".format(proxies))


def get_git_info(repo_path=None):
    """Retrieve information about git repository.

    If attempt fails, ``None`` will be returned.

    Args:
        repo_path (:obj:`str`, optional, default is ``None``):

            | Path to the repository from which extract information about git.
            | If ``None`` is passed, calling ``get_git_info`` is equivalent to calling
              ``git.Repo(search_parent_directories=True)``.
              Check `GitPython <https://gitpython.readthedocs.io/en/stable/reference.html#git.repo.base.Repo>`_
              docs for more information.

    Returns:
        :class:`~neptune.git_info.GitInfo` - An object representing information about git repository.

    Examples:

        .. code:: python3

            # Get git info from the current directory
            git_info = get_git_info('.')

    """
    try:
        # pylint:disable=bad-option-value,import-outside-toplevel
        import git

        repo = git.Repo(repo_path, search_parent_directories=True)

        commit = repo.head.commit

        return GitInfo(
            commit_id=commit.hexsha,
            message=commit.message,
            author_name=commit.author.name,
            author_email=commit.author.email,
            commit_date=commit.committed_datetime,
            repository_dirty=repo.is_dirty(untracked_files=True)
        )
    except:  # pylint: disable=bare-except
        return None


def with_api_exceptions_handler(func):
    def wrapper(*args, **kwargs):
        for retry in range(0, 11):
            try:
                return func(*args, **kwargs)
            except requests.exceptions.SSLError:
                raise SSLError()
            except (BravadoConnectionError, BravadoTimeoutError,
                    requests.exceptions.ConnectionError, requests.exceptions.Timeout,
                    HTTPRequestTimeout, HTTPServiceUnavailable, HTTPGatewayTimeout):
                if retry >= 6:
                    _logger.warning('Experiencing connection interruptions. Reestablishing communication with Neptune.')
                time.sleep(2 ** retry)
                continue
            except HTTPServerError:
                raise ServerError()
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code >= HTTPInternalServerError.status_code:
                    raise ServerError()
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                else:
                    raise
        raise ConnectionLost()

    return wrapper


def glob(pathname):
    # pylint: disable=unexpected-keyword-arg
    if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 5):
        return globlib.glob(pathname)
    else:
        return globlib.glob(pathname, recursive=True)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from pprint import pformat

from neptune.internal.storage.datastream import compress_to_tar_gz_in_memory, FileChunkStream


class UploadEntry(object):
    def __init__(self, source_path, target_path):
        self.source_path = source_path
        self.target_path = target_path

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

    def __hash__(self):
        """
        Returns the hash of source and target path
        """
        return hash((self.source_path, self.target_path))

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.__dict__)

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()


class UploadPackage(object):
    def __init__(self):
        self.items = []
        self.size = 0
        self.len = 0

    def reset(self):
        self.items = []
        self.size = 0
        self.len = 0

    def update(self, entry, size):
        self.items.append(entry)
        self.size += size
        self.len += 1

    def is_empty(self):
        return self.len == 0

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.__dict__)

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()


def scan_unique_upload_entries(upload_entries):
    """
    Returns upload entries for all files that could be found for given upload entries.
    In case of directory as upload entry, files we be taken from all subdirectories recursively.
    Any duplicated entries are removed.
    """
    walked_entries = set()
    for entry in upload_entries:
        if os.path.isdir(entry.source_path):
            for root, _, files in os.walk(entry.source_path):
                path_relative_to_entry_source = os.path.relpath(root, entry.source_path)
                target_root = os.path.normpath(os.path.join(entry.target_path, path_relative_to_entry_source))
                for filename in files:
                    walked_entries.add(UploadEntry(os.path.join(root, filename), os.path.join(target_root, filename)))
        else:
            walked_entries.add(entry)
    return walked_entries


def split_upload_files(upload_entries, max_package_size=1 * 1024 * 1024, max_files=500):
    current_package = UploadPackage()

    for entry in upload_entries:
        size = os.path.getsize(entry.source_path)

        if (size + current_package.size > max_package_size or current_package.len > max_files) \
                and not current_package.is_empty():
            yield current_package
            current_package.reset()
        current_package.update(entry, size)

    yield current_package


def normalize_file_name(name):
    return name.replace(os.sep, '/')


def upload_to_storage(upload_entries, upload_api_fun, upload_tar_api_fun, **kwargs):
    for package in split_upload_files(scan_unique_upload_entries(upload_entries)):
        if package.is_empty():
            break

        uploading_multiple_entries = package.len > 1
        creating_a_single_empty_dir = package.len == 1 and os.path.isdir(package.items[0].source_path)

        if uploading_multiple_entries or creating_a_single_empty_dir:
            data = compress_to_tar_gz_in_memory(upload_entries=package.items)
            upload_tar_api_fun(**dict(kwargs, data=data))
        else:
            file_chunk_stream = FileChunkStream(package.items[0])
            upload_api_fun(**dict(kwargs, data=file_chunk_stream))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from neptune.utils import validate_notebook_path


class Notebook(object):
    """It contains all the information about a Neptune Notebook

        Args:
            backend (:class:`~neptune.Backend`): A Backend object
            project (:class:`~neptune.projects.Project`): Project object
            _id (:obj:`str`): Notebook uuid
            owner (:obj:`str`): Creator of the notebook is the Notebook owner

        Examples:
            .. code:: python3

                # Create a notebook in Neptune.
                notebook = project.create_notebook('data_exploration.ipynb')

    """
    def __init__(self, backend, project, _id, owner):
        self._backend = backend
        self._project = project
        self._id = _id
        self._owner = owner

    @property
    def id(self):
        return self._id

    @property
    def owner(self):
        return self._owner

    def add_checkpoint(self, file_path):
        """Uploads new checkpoint of the notebook to Neptune

        Args:
            file_path (:obj:`str`): File path containing notebook contents

        Example:

            .. code:: python3

                # Create a notebook.
                notebook = project.create_notebook('file.ipynb')

                # Change content in your notebook & save it

                # Upload new checkpoint
                notebook.add_checkpoint('file.ipynb')
        """
        validate_notebook_path(file_path)

        with open(file_path) as f:
            return self._backend.create_checkpoint(self.id, os.path.abspath(file_path), f)

    def get_path(self):
        """Returns the path used to upload the current checkpoint of this notebook

        Returns:
            :obj:`str`: path of the current checkpoint
        """
        return self._backend.get_last_checkpoint(self._project, self._id).path

    def get_name(self):
        """Returns the name used to upload the current checkpoint of this notebook

        Returns:
            :obj:`str`: the name of current checkpoint
        """
        return self._backend.get_last_checkpoint(self._project, self._id).name



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

_logger = logging.getLogger(__name__)


class SystemMonitor(object):
    @staticmethod
    def requirements_installed():
        # pylint:disable=unused-import,unused-variable,bad-option-value,import-outside-toplevel
        try:
            import psutil
            return True
        except ImportError:
            _logger.warning('psutil is not installed. Hardware metrics will not be collected.')
            return False

    def cpu_count(self):
        return self._psutil().cpu_count()

    def cpu_percent(self):
        return self._psutil().cpu_percent()

    def virtual_memory(self):
        return self._psutil().virtual_memory()

    @classmethod
    def _psutil(cls):
        # pylint:disable=bad-option-value,import-outside-toplevel
        import psutil
        return psutil



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import random

from websocket import WebSocketConnectionClosedException, WebSocketTimeoutException

from neptune.internal.websockets.websocket_client_adapter import WebsocketClientAdapter, WebsocketNotConnectedException


class ReconnectingWebsocket(object):
    def __init__(self, url, oauth2_session, shutdown_event):
        self.url = url
        self.client = WebsocketClientAdapter()
        self._shutdown_event = shutdown_event
        self._oauth2_session = oauth2_session
        self._reconnect_counter = ReconnectCounter()
        self._token = oauth2_session.token

    def shutdown(self):
        self._shutdown_event.set()
        self.client.abort()
        self.client.shutdown()

    def recv(self):
        if not self.client.connected:
            self._try_to_establish_connection()
        while self._is_active():
            try:
                data = self.client.recv()
                self._on_successful_connect()
                return data
            except WebSocketTimeoutException:
                raise
            except WebSocketConnectionClosedException:
                if self._is_active():
                    self._handle_lost_connection()
                else:
                    raise
            except WebsocketNotConnectedException:
                if self._is_active():
                    self._handle_lost_connection()
            except Exception:
                if self._is_active():
                    self._handle_lost_connection()

    def _is_active(self):
        return not self._shutdown_event.is_set()

    def _on_successful_connect(self):
        self._reconnect_counter.clear()

    def _try_to_establish_connection(self):
        try:
            self._request_token_refresh()
            if self.client.connected:
                self.client.shutdown()
            self.client.connect(url=self.url, token=self._token)
        except Exception:
            self._shutdown_event.wait(self._reconnect_counter.calculate_delay())

    def _handle_lost_connection(self):
        self._reconnect_counter.increment()
        self._try_to_establish_connection()

    def _request_token_refresh(self):
        self._token = self._oauth2_session.refresh_token(token_url=self._oauth2_session.auto_refresh_url)


class ReconnectCounter(object):
    MAX_RETRY_DELAY = 128

    def __init__(self):
        self.retries = 0

    def clear(self):
        self.retries = 0

    def increment(self):
        self.retries += 1

    def calculate_delay(self):
        return self._compute_delay(self.retries, self.MAX_RETRY_DELAY)

    @classmethod
    def _compute_delay(cls, attempt, max_delay):
        delay = cls._full_jitter_delay(attempt, max_delay)
        return delay

    @classmethod
    def _full_jitter_delay(cls, attempt, cap):
        exp = min(2 ** (attempt - 1), cap)
        return random.uniform(0, exp)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import unittest

from mock import MagicMock

from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.gpu.gpu_monitor import GPUMonitor
from neptune.internal.hardware.resources.system_resource_info_factory import SystemResourceInfoFactory
from neptune.internal.hardware.system.system_monitor import SystemMonitor
from neptune.utils import IS_WINDOWS
from tests.neptune.utils.assertions import AssertionExtensions


class TestSystemResourceInfoFactoryIntegration(unittest.TestCase, AssertionExtensions):
    def test_whole_system_resource_info(self):
        # given
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=GPUMonitor(), os_environ=os.environ)

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assert_float_greater_than(resource_info.cpu_core_count, 0)
        self.assert_int_greater_than(resource_info.memory_amount_bytes, 0)
        self.assert_int_greater_or_equal(resource_info.gpu_card_count, 0)
        self.assert_int_greater_or_equal(resource_info.gpu_memory_amount_bytes, 0)

    @unittest.skipIf(IS_WINDOWS, "Windows doesn't have cgroups")
    def test_cgroup_resource_info(self):
        # given
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=GPUMonitor(), os_environ=os.environ)

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.CGROUP)

        # then
        self.assert_float_greater_than(resource_info.cpu_core_count, 0)
        self.assert_int_greater_than(resource_info.memory_amount_bytes, 0)
        self.assert_int_greater_or_equal(resource_info.gpu_card_count, 0)
        self.assert_int_greater_or_equal(resource_info.gpu_memory_amount_bytes, 0)


class TestSystemResourceInfoFactory(unittest.TestCase):
    def test_gpu_card_indices_without_cuda_env_variable(self):
        # given
        gpu_monitor = MagicMock(spec_set=GPUMonitor)
        gpu_monitor.get_card_count.return_value = 2
        # and
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=gpu_monitor, os_environ=dict())

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assertEqual([0, 1], resource_info.gpu_card_indices)

    def test_gpu_card_indices_based_on_cuda_env_variable(self):
        # given
        gpu_monitor = MagicMock(spec_set=GPUMonitor)
        gpu_monitor.get_card_count.return_value = 4
        # and
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=gpu_monitor, os_environ={u'CUDA_VISIBLE_DEVICES': u'1,3'})

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assertEqual([1, 3], resource_info.gpu_card_indices)

    def test_empty_gpu_card_indices_on_cuda_env_variable_minus_one(self):
        # given
        gpu_monitor = MagicMock(spec_set=GPUMonitor)
        gpu_monitor.get_card_count.return_value = 4
        # and
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=gpu_monitor, os_environ={u'CUDA_VISIBLE_DEVICES': u'-1'})

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assertEqual([], resource_info.gpu_card_indices)

    def test_should_ignore_gpu_indices_after_index_out_of_range(self):
        # given
        gpu_monitor = MagicMock(spec_set=GPUMonitor)
        gpu_monitor.get_card_count.return_value = 4
        # and
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=gpu_monitor, os_environ={u'CUDA_VISIBLE_DEVICES': u'1,3,5,2'})

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assertEqual([1, 3], resource_info.gpu_card_indices)

    def test_ignore_empty_cuda_env_variable(self):
        # given
        gpu_monitor = MagicMock(spec_set=GPUMonitor)
        gpu_monitor.get_card_count.return_value = 2
        # and
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=gpu_monitor, os_environ={u'CUDA_VISIBLE_DEVICES': u''})

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assertEqual([0, 1], resource_info.gpu_card_indices)

    def test_should_ignore_invalid_cuda_env_variable_syntax(self):
        # given
        gpu_monitor = MagicMock(spec_set=GPUMonitor)
        gpu_monitor.get_card_count.return_value = 4
        # and
        system_resource_info_factory = SystemResourceInfoFactory(
            system_monitor=SystemMonitor(), gpu_monitor=gpu_monitor, os_environ={u'CUDA_VISIBLE_DEVICES': u'1,3,abc'})

        # when
        resource_info = system_resource_info_factory.create(GaugeMode.SYSTEM)

        # then
        self.assertEqual([0, 1, 2, 3], resource_info.gpu_card_indices)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

_logger = logging.getLogger(__name__)


class CustomAbortImpl(object):
    def __init__(self, runnable):
        self.__runnable = runnable

    def abort(self):
        self.__runnable()


class DefaultAbortImpl(object):
    KILL_TIMEOUT = 5

    def __init__(self, pid):
        self._pid = pid

    @staticmethod
    def requirements_installed():
        # pylint:disable=unused-import,unused-variable,bad-option-value,import-outside-toplevel
        try:
            import psutil
            return True
        except ImportError:
            _logger.warning('psutil is not installed. The experiment will not be abortable.')
            return False

    def abort(self):
        # pylint:disable=bad-option-value,import-outside-toplevel
        import psutil

        process = None
        try:
            process = psutil.Process(self._pid)
        except psutil.NoSuchProcess:
            pass

        if process is not None:
            processes = self._get_processes(process)
            for p in processes:
                self._abort(p)
            _, alive = psutil.wait_procs(processes, timeout=self.KILL_TIMEOUT)
            for p in alive:
                self._kill(p)

    @staticmethod
    def _get_processes(process):
        # pylint:disable=bad-option-value,import-outside-toplevel
        import psutil

        try:
            return [process] + process.children(recursive=True)
        except psutil.NoSuchProcess:
            return []

    def _abort(self, process):
        # pylint:disable=bad-option-value,import-outside-toplevel
        import psutil

        try:
            process.terminate()
        except psutil.NoSuchProcess:
            pass

    def _kill(self, process):
        # pylint:disable=bad-option-value,import-outside-toplevel
        import psutil

        for process in self._get_processes(process):
            try:
                if process.is_running():
                    process.kill()
            except psutil.NoSuchProcess:
                pass



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import unicode_literals

import time
import unittest

import mock

from neptune.internal.channels.channels import ChannelIdWithValues, ChannelValue, ChannelType, ChannelNamespace
from neptune.internal.channels.channels_values_sender import ChannelsValuesSender, ChannelsValuesSendingThread
from tests.neptune.experiments_object_factory import a_channel


class TestChannelsValuesSender(unittest.TestCase):
    _TS = time.time()

    _EXPERIMENT = mock.MagicMock()

    _NUMERIC_CHANNEL = a_channel()
    _NUMERIC_CHANNEL.update(name=u'1numeric', channelType=ChannelType.NUMERIC.value)

    _TEXT_CHANNEL = a_channel()
    _TEXT_CHANNEL.update(name=u'2text', channelType=ChannelType.TEXT.value)

    _IMAGE_CHANNEL = a_channel()
    _IMAGE_CHANNEL.update(name=u'3image', channelType=ChannelType.IMAGE.value)

    _OTHER_CHANNEL = a_channel()

    _CHANNELS = {c.name: c for c in [_NUMERIC_CHANNEL, _TEXT_CHANNEL, _IMAGE_CHANNEL, _OTHER_CHANNEL]}

    # pylint: disable=protected-access
    _BATCH_SIZE = ChannelsValuesSendingThread._MAX_VALUES_BATCH_LENGTH
    # pylint: disable=protected-access
    _IMAGES_BATCH_SIZE = ChannelsValuesSendingThread._MAX_IMAGE_VALUES_BATCH_LENGTH

    def setUp(self):
        # pylint: disable=protected-access
        self._EXPERIMENT._get_channels.return_value = self._CHANNELS

        # pylint: disable=unused-argument
        def create_channel_fun(channel_name, channel_type, channel_namespace=ChannelNamespace.USER):
            if channel_name == self._NUMERIC_CHANNEL.name:
                return self._NUMERIC_CHANNEL
            if channel_name == self._TEXT_CHANNEL.name:
                return self._TEXT_CHANNEL
            if channel_name == self._IMAGE_CHANNEL.name:
                return self._IMAGE_CHANNEL
            if channel_name == self._OTHER_CHANNEL.name:
                return self._OTHER_CHANNEL
            raise ValueError("unexpected channel")

        self._EXPERIMENT._create_channel = mock.MagicMock(side_effect=create_channel_fun)

    def tearDown(self):
        self._EXPERIMENT.reset_mock()

    def test_send_values_on_join(self):
        # given
        channel_value = ChannelValue(x=1, y="value", ts=self._TS)
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        channels_values_sender.send(self._TEXT_CHANNEL.name, self._TEXT_CHANNEL.channelType, channel_value)
        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        self._EXPERIMENT._send_channels_values.assert_called_with([ChannelIdWithValues(
            channel_id=self._TEXT_CHANNEL.id,
            channel_values=[channel_value]
        )])

    def test_send_values_in_multiple_batches(self):
        # given
        channels_values = [ChannelValue(x=i, y="value{}".format(i), ts=self._TS + i)
                           for i in range(0, self._BATCH_SIZE * 3)]
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in channels_values:
            channels_values_sender.send(self._TEXT_CHANNEL.name, self._TEXT_CHANNEL.channelType, channel_value)
        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        self.assertEqual(self._EXPERIMENT._send_channels_values.mock_calls, [
            mock.call._send_channels_values([ChannelIdWithValues(
                channel_id=self._TEXT_CHANNEL.id,
                channel_values=channels_values[0:self._BATCH_SIZE]
            )]),
            mock.call._send_channels_values([ChannelIdWithValues(
                channel_id=self._TEXT_CHANNEL.id,
                channel_values=channels_values[self._BATCH_SIZE:self._BATCH_SIZE * 2]
            )]),
            mock.call._send_channels_values([ChannelIdWithValues(
                channel_id=self._TEXT_CHANNEL.id,
                channel_values=channels_values[self._BATCH_SIZE * 2:self._BATCH_SIZE * 3]
            )])
        ])

    def test_send_images_in_smaller_batches(self):
        # and
        channels_values = [ChannelValue(x=i, y="base64Image==", ts=self._TS + i)
                           for i in range(0, self._IMAGES_BATCH_SIZE * 3)]
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in channels_values:
            channels_values_sender.send(self._IMAGE_CHANNEL.name, self._IMAGE_CHANNEL.channelType, channel_value)
        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        self.assertEqual(self._EXPERIMENT._send_channels_values.mock_calls, [
            mock.call._send_channels_values([ChannelIdWithValues(
                channel_id=self._IMAGE_CHANNEL.id,
                channel_values=channels_values[0:self._IMAGES_BATCH_SIZE]
            )]),
            mock.call._send_channels_values([ChannelIdWithValues(
                channel_id=self._IMAGE_CHANNEL.id,
                channel_values=channels_values[self._IMAGES_BATCH_SIZE:self._IMAGES_BATCH_SIZE * 2]
            )]),
            mock.call._send_channels_values([ChannelIdWithValues(
                channel_id=self._IMAGE_CHANNEL.id,
                channel_values=channels_values[self._IMAGES_BATCH_SIZE * 2:self._IMAGES_BATCH_SIZE * 3]
            )])
        ])

    def test_send_values_from_multiple_channels(self):
        # given
        numeric_values = [ChannelValue(x=i, y=i, ts=self._TS + i)
                          for i in range(0, 3)]

        text_values = [ChannelValue(x=i, y="text", ts=self._TS + i)
                       for i in range(0, 3)]

        image_values = [ChannelValue(x=i, y="base64Image==", ts=self._TS + i)
                        for i in range(0, 3)]
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in numeric_values:
            channels_values_sender.send(self._NUMERIC_CHANNEL.name, self._NUMERIC_CHANNEL.channelType, channel_value)

        for channel_value in text_values:
            channels_values_sender.send(self._TEXT_CHANNEL.name, self._TEXT_CHANNEL.channelType, channel_value)

        for channel_value in image_values:
            channels_values_sender.send(self._IMAGE_CHANNEL.name, self._IMAGE_CHANNEL.channelType, channel_value)

        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        (args, _) = self._EXPERIMENT._send_channels_values.call_args
        self.assertEqual(len(args), 1)
        self.assertEqual(sorted(args[0]), sorted([
            ChannelIdWithValues(
                channel_id=self._NUMERIC_CHANNEL.id,
                channel_values=numeric_values
            ),
            ChannelIdWithValues(
                channel_id=self._TEXT_CHANNEL.id,
                channel_values=text_values
            ),
            ChannelIdWithValues(
                channel_id=self._IMAGE_CHANNEL.id,
                channel_values=image_values
            )]))

    __TIMEOUT = 0.1

    @mock.patch('neptune.internal.channels.channels_values_sender.ChannelsValuesSendingThread._SLEEP_TIME', __TIMEOUT)
    def test_send_when_waiting_for_next_value_timed_out(self):
        # given
        numeric_values = [ChannelValue(x=i, y=i, ts=self._TS + i)
                          for i in range(0, 3)]

        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in numeric_values:
            channels_values_sender.send(self._NUMERIC_CHANNEL.name, self._NUMERIC_CHANNEL.channelType, channel_value)

        # and
        time.sleep(self.__TIMEOUT * 2)

        # then
        # pylint: disable=protected-access
        self._EXPERIMENT._send_channels_values.assert_called_with([ChannelIdWithValues(
            channel_id=self._NUMERIC_CHANNEL.id,
            channel_values=numeric_values
        )])

        # and
        self._EXPERIMENT._send_channels_values.reset_mock()
        channels_values_sender.join()
        # and
        self._EXPERIMENT._send_channels_values.assert_not_called()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.internal.hardware.gauges.cpu import CGroupCpuUsageGauge, SystemCpuUsageGauge
from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.gauges.gpu import GpuMemoryGauge, GpuUsageGauge
from neptune.internal.hardware.gauges.memory import CGroupMemoryUsageGauge, SystemMemoryUsageGauge


class GaugeFactory(object):
    def __init__(self, gauge_mode):
        self.__gauge_mode = gauge_mode

    def create_cpu_usage_gauge(self):
        if self.__gauge_mode == GaugeMode.SYSTEM:
            return SystemCpuUsageGauge()
        elif self.__gauge_mode == GaugeMode.CGROUP:
            return CGroupCpuUsageGauge()
        else:
            raise self.__invalid_gauge_mode_exception()

    def create_memory_usage_gauge(self):
        if self.__gauge_mode == GaugeMode.SYSTEM:
            return SystemMemoryUsageGauge()
        elif self.__gauge_mode == GaugeMode.CGROUP:
            return CGroupMemoryUsageGauge()
        else:
            raise self.__invalid_gauge_mode_exception()

    @staticmethod
    def create_gpu_usage_gauge(card_index):
        return GpuUsageGauge(card_index=card_index)

    @staticmethod
    def create_gpu_memory_gauge(card_index):
        return GpuMemoryGauge(card_index=card_index)

    def __invalid_gauge_mode_exception(self):
        return ValueError(str(u'Invalid gauge mode: {}'.format(self.__gauge_mode)))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time

import jwt
from bravado.requests_client import Authenticator
from oauthlib.oauth2 import TokenExpiredError
from requests.auth import AuthBase
from requests_oauthlib import OAuth2Session

from neptune.utils import with_api_exceptions_handler, update_session_proxies


class NeptuneAuth(AuthBase):

    def __init__(self, session):
        self.session = session
        self.token_expires_at = 0

    def __call__(self, r):
        try:
            return self._add_token(r)
        except TokenExpiredError:
            self._refresh_token()
            return self._add_token(r)

    def _add_token(self, r):
        # pylint: disable=protected-access
        r.url, r.headers, r.body = self.session._client.add_token(r.url,
                                                                  http_method=r.method,
                                                                  body=r.body,
                                                                  headers=r.headers)
        return r

    @with_api_exceptions_handler
    def refresh_token_if_needed(self):
        if self.token_expires_at - time.time() < 30:
            self._refresh_token()

    def _refresh_token(self):
        self.session.refresh_token(self.session.auto_refresh_url)
        if self.session.token is not None and self.session.token.get('access_token') is not None:
            decoded_json_token = jwt.decode(self.session.token.get('access_token'), verify=False)
            self.token_expires_at = decoded_json_token.get(u'exp')


class NeptuneAuthenticator(Authenticator):

    def __init__(self, auth_tokens, ssl_verify, proxies):
        super(NeptuneAuthenticator, self).__init__(host='')
        decoded_json_token = jwt.decode(auth_tokens.accessToken, verify=False)
        expires_at = decoded_json_token.get(u'exp')
        client_name = decoded_json_token.get(u'azp')
        refresh_url = u'{realm_url}/protocol/openid-connect/token'.format(realm_url=decoded_json_token.get(u'iss'))
        token = {
            u'access_token': auth_tokens.accessToken,
            u'refresh_token': auth_tokens.refreshToken,
            u'expires_in': expires_at - time.time()
        }
        session = OAuth2Session(
            client_id=client_name,
            token=token,
            auto_refresh_url=refresh_url,
            auto_refresh_kwargs={'client_id': client_name},
            token_updater=_no_token_updater
        )
        session.verify = ssl_verify

        update_session_proxies(session, proxies)

        self.auth = NeptuneAuth(session)

    def matches(self, url):
        return True

    def apply(self, request):
        self.auth.refresh_token_if_needed()
        request.auth = self.auth
        return request


def _no_token_updater():
    # For unit tests.
    return None



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import io
import os

from PIL import Image
import numpy
import six

from neptune.exceptions import FileNotFound, InvalidChannelValue


def get_image_content(image):
    if isinstance(image, six.string_types):
        if not os.path.exists(image):
            raise FileNotFound(image)
        with open(image, 'rb') as image_file:
            return image_file.read()

    elif isinstance(image, numpy.ndarray):
        shape = image.shape
        if len(shape) == 2:
            return _get_pil_image_data(Image.fromarray(image.astype(numpy.uint8)))
        if len(shape) == 3:
            if shape[2] == 1:
                array2d = numpy.array([[col[0] for col in row] for row in image])
                return _get_pil_image_data(Image.fromarray(array2d.astype(numpy.uint8)))
            if shape[2] in (3, 4):
                return _get_pil_image_data(Image.fromarray(image.astype(numpy.uint8)))
        raise ValueError("Incorrect size of numpy.ndarray. Should be 2-dimensional or"
                         "3-dimensional with 3rd dimension of size 1, 3 or 4.")

    elif isinstance(image, Image.Image):
        return _get_pil_image_data(image)

    else:
        try:
            from matplotlib import figure
            if isinstance(image, figure.Figure):
                return _get_pil_image_data(_figure_to_pil_image(image))
        except ImportError:
            pass

    raise InvalidChannelValue(expected_type='image', actual_type=type(image).__name__)


def _get_pil_image_data(image):
    with io.BytesIO() as image_buffer:
        image.save(image_buffer, format='PNG')
        return image_buffer.getvalue()


def _figure_to_pil_image(figure):
    figure.canvas.draw()
    return Image.frombytes('RGB', figure.canvas.get_width_height(), figure.canvas.tostring_rgb())
